/*
 *  Copyright (c) 1992 John E. Davis  (davis@amy.tch.harvard.edu)
 *  All Rights Reserved.
 * Munged by Dan Kegel (dank@blacks.jpl.nasa.gov) for use in uwho.
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <sgtty.h>
#define TTY_DESCR 0

#define KeyBoard_Quit 0		/* should be global set during SIGINT */

/*--------------------------------------------------------------------------
 Call with 1 to install, 0 to deinstall raw mode
--------------------------------------------------------------------------*/
void raw_set_stdio(init)
    int init;
{
    struct sgttyb  newtty;
    struct tchars tc;
    struct ltchars lc;
    static struct ltchars  OLDTTY_LC;
    static struct tchars   OLDTTY_TC;
    static struct sgttyb   OLDTTY;

    if (init) {
	ioctl(TTY_DESCR, TIOCGLTC,&OLDTTY_LC);
	ioctl(TTY_DESCR, TIOCGETC,&OLDTTY_TC);
	ioctl(TTY_DESCR, TIOCGETP, &OLDTTY);
	newtty = OLDTTY;
	newtty.sg_flags |= CBREAK;
	newtty.sg_flags &= ~(ECHO);
#if 0
	newtty.sg_erase = 0;
	newtty.sg_kill = 0;
	tc.t_intrc = 0;
	tc.t_quitc = 0;
	tc.t_eofc = 0;
	tc.t_brkc = 0;
	lc.t_suspc = 0;
	lc.t_dsuspc = 0;
#endif
	ioctl(TTY_DESCR, TIOCSETP, &newtty);
	ioctl(TTY_DESCR, TIOCSLTC,&lc);
	ioctl(TTY_DESCR, TIOCSETC,&tc);
    } else {
	ioctl(TTY_DESCR, TIOCSETP, &OLDTTY);
	ioctl(TTY_DESCR, TIOCSLTC, &OLDTTY_LC);
	ioctl(TTY_DESCR, TIOCSETC, &OLDTTY_TC);
    }
}

/*--------------------------------------------------------------------------
 Return -1 if no key, 3 if user hit interrupt key, otherwise return key.
--------------------------------------------------------------------------*/
int raw_getc()
{
    char c;

    if (!KeyBoard_Quit && !raw_kbhit())
	return -1;

    if (!KeyBoard_Quit && read(TTY_DESCR, &c, 1) < 0) {
	/* Gack!  Read failed!  Pretend interrupt hit. */
	c = 3;
    }
    if (KeyBoard_Quit) c = 3;
    return(c);
}

/*--------------------------------------------------------------------------
 Return 0 if no key ready, 1 if key ready.
--------------------------------------------------------------------------*/
int raw_kbhit()
{
    struct timeval wait;
    int readfd, writefd, exceptfd, ret;
    long usecs = 0;
    int secs = 0;

    /* extern int select(int, int *, int *, int *, struct timeval *); */

    if (secs == 0) usecs = 100000;  /* 1/10th of a sec */
    wait.tv_sec = secs;
    wait.tv_usec = usecs;

    readfd = 1 << TTY_DESCR;
    writefd = exceptfd = 0;

    ret = select(TTY_DESCR + 1, &readfd, &writefd, &exceptfd, &wait);
    return(ret);
}

/*--------------------------------------------------------------------------
 Get the size of the terminal
--------------------------------------------------------------------------*/
void raw_get_term_dimensions(cols, rows)
    int *cols;
    int *rows;
{
    struct winsize wind_struct;
      
   ioctl(2,TIOCGWINSZ,&wind_struct);
   *cols = (int) wind_struct.ws_col;
   *rows = (int) wind_struct.ws_row;
   if (*rows <= 0) *rows = 24;
   if (*cols <= 0) *cols = 80;
}

